/**
 * Title:        Viterbi Decoder for the Voyager code
 * Description:  A Viterbi decoding program for the Voyager code (2,1,6,10)
 * Copyright:    Copyright (c) 2001, Ling Li
 * Company:      California Institute of Technology
 * @author Ling Li
 * @email  ling@cs.caltech.edu
 * @version 1.0
 */

package ecc;

public class Voyager {
    final int states = 64;
    final int state_mask = states-1;

    // trellis[st][x] is the next state from st with input x.
    private int trellis[][];
    // outputs(st, x) is the output at state st with input x
    private int output0[][], output1[][];

    private void make_trellis() {
        for (int st = 0; st < states; st++) {
            int newst = (st << 1) & state_mask;
            for (int bit = 0; bit < 2; bit++) {
                trellis[st][bit] = newst | bit;
                // st >> i is D^(i+1)
                output0[st][bit] = 1 & (bit ^ (st >> 1) ^ (st >> 2)
                                            ^ (st >> 4) ^ (st >> 5));
                output1[st][bit] = 1 & (bit ^ st ^ (st >> 1)
                                            ^ (st >> 2) ^ (st >> 5));
            }
        }
    }

    public Voyager() {
        trellis = new int[states][2];
        output0 = new int[states][2];
        output1 = new int[states][2];

        make_trellis();
    }

    /** y = encode(x);
     *  where x is a row vector containing information
     *  we will not ensure the state to be back to 0.
     */
    public int[][] encode(int[] x) {
        int len = x.length;
        int y[][], i;

        y = new int[len][2];

       // if (len < 6)

        y[0][0] = y[0][1] = x[0];

        y[1][0] = x[1];
        y[1][1] = x[1] ^ x[0];

        y[2][0] = x[2] ^ x[0];
        y[2][1] = x[2] ^ x[1] ^ x[0];

        y[3][0] = x[3] ^ x[1] ^ x[0];
        y[3][1] = x[3] ^ x[2] ^ x[1] ^ x[0];

        y[4][0] = x[4] ^ x[2] ^ x[1];
        y[4][1] = x[4] ^ x[3] ^ x[2] ^ x[1];

        y[5][0] = x[5] ^ x[3] ^ x[2] ^ x[0];
        y[5][1] = x[5] ^ x[4] ^ x[3] ^ x[2];

        for (i = 6; i < len; i++) {
            y[i][0] = x[i] ^ x[i-2] ^ x[i-3] ^ x[i-5] ^ x[i-6];
            y[i][1] = x[i] ^ x[i-1] ^ x[i-2] ^ x[i-3] ^ x[i-6];
        }

        return y;
    }

    public int[] decode(double[][] r, int trunc_len) {
        int len = r.length;
        if (len <= trunc_len) return null;
        int x[] = new int[len - trunc_len];

        /*  In every stage, we need to save the truncated surviors.
         *  and the metrics sum (we want to minimize it)
         */
        long curr_x[] = new long[states];
        long next_x[] = new long[states];
        double curr_sum[] = new double[states];
        double next_sum[] = new double[states];
        int min_idx, st, i;

        curr_sum[0] = 0;    // only state 000000 is valid
        curr_x[0] = 0;
        for (st = 1; st < states; st++)
            curr_sum[st] = Double.POSITIVE_INFINITY;
        min_idx = 0;

        for (i = 0; i < len; i++) {
            if (i >= trunc_len) {
                x[i-trunc_len] = (int) (curr_x[min_idx] >>> (trunc_len-1)) & 1;
            }

            for (st = 0; st < states; st++)
                next_sum[st] = Double.POSITIVE_INFINITY;

            // update all states
            for (st = 0; st < states; st++) {
                for (int bit = 0; bit < 2; bit++) {
                    /* dist =  - (1-2output0) * r0 - (1-2output1) * r1; */
                    double newsum = curr_sum[st]
                                      - (1 - 2 * output0[st][bit]) * r[i][0]
                                      - (1 - 2 * output1[st][bit]) * r[i][1];

                    int newst = trellis[st][bit];

                    if (next_sum[newst] > newsum) {
                        next_sum[newst] = newsum;
                        next_x[newst] = (curr_x[st] << 1) | bit;
                        if (newsum < next_sum[min_idx])
                            min_idx = newst;
                    }
                }
            }

            long temp_x[] = curr_x;
            curr_x = next_x; next_x = temp_x;
            double temp_sum[] = curr_sum;
            curr_sum = next_sum; next_sum = temp_sum;
        }
        return x;
    }

    public int[] decode(int[][] r, int trunc_len) {
        int len = r.length;
        if (len <= trunc_len) return (new int[0]);
        int x[] = new int[len - trunc_len];

        /*  In every stage, we need to save the truncated surviors.
         *  and the metrics sum (we want to minimize it)
         */
        long curr_x[] = new long[states];
        long next_x[] = new long[states];
        int curr_sum[] = new int[states];
        int next_sum[] = new int[states];
        int min_idx, st, i;

        curr_sum[0] = 0;    // only state 000000 is valid
        curr_x[0] = 0;
        for (st = 1; st < states; st++)
            curr_sum[st] = Integer.MAX_VALUE;
        min_idx = 0;

        for (i = 0; i < len; i++) {
            if (i >= trunc_len) {
                x[i-trunc_len] = (int) (curr_x[min_idx] >>> (trunc_len-1)) & 1;
            }

            for (st = 0; st < states; st++)
                next_sum[st] = Integer.MAX_VALUE;

            // update all states
            for (st = 0; st < states; st++) {
                for (int bit = 0; bit < 2; bit++) {
                    int newsum = curr_sum[st];
                    if (output0[st][bit] != r[i][0])
                         newsum++;
                    if (output1[st][bit] != r[i][1])
                         newsum++;

                    int newst = trellis[st][bit];

                    if (next_sum[newst] > newsum) {
                        next_sum[newst] = newsum;
                        next_x[newst] = (curr_x[st] << 1) | bit;
                        if (newsum < next_sum[min_idx])
                            min_idx = newst;
                    }
                }
            }

            long temp_x[] = curr_x;
            curr_x = next_x; next_x = temp_x;
            int temp_sum[] = curr_sum;
            curr_sum = next_sum; next_sum = temp_sum;
        }
        return x;
    }
}